/* MinWin Application structure
 *
 * An Application encapsulates the global state provided by
 * the OS when starting the application and it manages the event queue.
 *
 * Written by Ben Hinkle and released to the public domain, as
 * explained at http://creativecommons.org/licenses/publicdomain
 * Email comments and bug reports to ben.hinkle@gmail.com
 */

module minwin.app;

private {
  import minwin.logging;
  import minwin.multidg;
  import minwin.event;
  import std.string;
}

extern (C) void gc_init();
extern (C) void gc_term();
extern (C) void _minit();
extern (C) void _moduleCtor();
extern (C) void _moduleUnitTests();
extern (C) int MinWinMain(Application* app);


const int NO_ID = -1;
// Exception raised when a system call fails.
class MinWinException : Exception {
  int id = NO_ID;
  this(char[] msg) { 
    super(msg); 
  }
  this(char[] msg, int id) { 
    this.id = id;
    if (id == NO_ID)
      super(msg);
    else
      super(format("%s (error code %x)",msg,id));
  }
}

version (Windows) {
} else {
  // Assert a condition is true and if not throw an exception 
  void sysAssert(bool passed, char[] msg, int id = 0) {
    if (!passed) {
      throw new MinWinException(msg,id);
    }
  }
}

Application gApp;

version (Windows) {
  
  private import minwin.mswindows;

  // Assert a condition is true and if not throw an exception 
  void sysAssert(bool passed, char[] msg, int id = 0) {
    if (!passed) {
      if (GetVersion() < 0x80_00_00_00 || id != 0) {
	throw new MinWinException(msg,id);
      } else {
	throw new MinWinException(msg,GetLastError());
      }
    }
  }
  
  extern (Windows)
  void MinWinIdleProc(HWND hWnd, uint msg, uint id, DWORD time) {
    gApp.idleDelegate();
  }

  struct Application {

    // lookup a string resource
    char[] rsrc(char[] id, char[] group = "strings") {
      char[] res;
      res.length = 64;
      int len;
      for (;;) {
	int len = cast(int)LoadStringA(gApp.hInstance,toStringz(id),
				       res.ptr,res.length);
	if (len < res.length)
	  break;
	else
	  res.length = 2*res.length;
      }
      return res[0..len];
    }

    char[][] cmdLineArgs(){
      // TODO: parse cmd line args
      return parsedCmdLineArgs;
    }
    
    MultiDelegate!() idleDelegate;
    uint fTimerID;

    uint fIdleTime;
    uint idleTime() {
      return fIdleTime;
    }
    void idleTime(uint t) { // time 0 means stop
      fIdleTime = t;
      if (fTimerID) {
	BOOL ok = KillTimer(null,fTimerID);
	sysAssert(ok != false, "Failed to kill existing timer");
      }
      if (t) {
	fTimerID = SetTimer(null,0,t,&MinWinIdleProc);
	sysAssert(fTimerID != 0, "Failed to set timer");
      }
    }

    // begins processing events until exitEventLoop is called or
    // the OS posts a Quit message
    int enterEventLoop() {
      while (nextEvent(&event)) {
	// TODO: check IsDialogMessage 
	dispatchEvent(&event);
      }
      return 1;
    }

    void exitEventLoop() {
      PostQuitMessage(0);
    }

    HINSTANCE hInstance;
    HINSTANCE hPrevInstance;
    LPSTR lpCmdLine;
    char[][] parsedCmdLineArgs;
    int nCmdShow;
    Event event;
  }

  extern (Windows)
    int WinMain(HINSTANCE hInstance,
		HINSTANCE hPrevInstance,
		LPSTR lpCmdLine,
		int nCmdShow) {
      int result;
      gc_init();			// initialize garbage collector
      _minit();			// initialize module constructor table

      try
	{
	  _moduleCtor();		// call module constructors
	  _moduleUnitTests();	// run unit tests (optional)

	  gApp.hInstance = hInstance;
	  gApp.hPrevInstance = hPrevInstance;
	  gApp.nCmdShow = nCmdShow;
	  version (LOG) log.writefln("about to enter MinWinMain");
	  result = MinWinMain(&gApp);
	  version (LOG) log.writefln("done with MinWinMain");
	}

      catch (Object o)		// catch any uncaught exceptions
	{
	  MessageBoxA(null, o.toString().ptr, "Error",
		      MB_OK | MB_ICONEXCLAMATION);
	  result = 0;		// failed
	}
      version(LOG) log.close();
      if (gApp.fTimerID) {
	BOOL ok = KillTimer(null,gApp.fTimerID);
	sysAssert(ok != false, "Failed to kill existing timer at exit");
      }
      gc_term();			// run finalizers; terminate garbage collector
      return result;
  }

} else version (Motif) {

  private {
    import minwin.motif;
    import std.string;
    import std.file;
  }

  extern (C)
  void MinWinIdleProc(XtPointer closure,XtIntervalId* id) {
    gApp.idleDelegate();
    gApp.idleTime = gApp.idleTime; // reset timer
  }

  struct Application {

    char[][] cmdLineArgs;

    // lookup a string resource
    char[] rsrc(char[] id, char[] group = "strings") {
      return "";
    }

    char[] resourcePath;

    MultiDelegate!() idleDelegate;
    uint fIdleTime;
    uint idleTime() {
      return fIdleTime;
    }

    XtIntervalId fTimerID;

    static 
    void idleTime(uint t) {
      gApp.fIdleTime = t;
      if (gApp.fTimerID)
	XtRemoveTimeOut(gApp.fTimerID);
      if (t)
	gApp.fTimerID = XtAppAddTimeOut(gApp.context,t,
				       &MinWinIdleProc,null);
    }

    int enterEventLoop() {
      XtAppMainLoop(context);
      return 1;
    }

    void exitEventLoop() {
      XtAppSetExitFlag(context);
    }

    XtAppContext context;
    Display* display;
    ImlibData* imlibdata;

  }

  int main(char[][] args) {
    gApp.cmdLineArgs = args;

    XtToolkitInitialize();
    gApp.context = XtCreateApplicationContext();
    char*[] argv;
    argv.length = args.length;
    foreach (int n, char[] str; args) {
      argv[n] = toStringz(str);
    }
    int argc = args.length;
    gApp.display = XtOpenDisplay(gApp.context,null,"test","appclass",null,0,
				 &argc,argv.ptr);
    gApp.imlibdata = Imlib_init(gApp.display);

    int res = MinWinMain(&gApp);
    version(LOG) log.close();
    return res;
  }

} else version (GTK) {

  private {
    import minwin.gtk;
    import std.string;
    import std.file;
  }

  extern (C)
  gboolean MinWinIdleProc(gpointer data) {
    gApp.idleDelegate();
    return true;
  }

  struct Application {

    char[][] cmdLineArgs;

    // lookup a string resource
    char[] rsrc(char[] id, char[] group = "strings") {
      // TODO
      return "";
    }

    char[] resourcePath;

    int enterEventLoop() {
      gtk_main();
      return 1;
    }

    void exitEventLoop() {
      gtk_main_quit();
    }

    MultiDelegate!() idleDelegate;
    uint fIdleTime;
    uint idleTime() {
      return fIdleTime;
    }

    guint fTimerID;

    static 
    void idleTime(uint t) {
      gApp.fIdleTime = t;
      if (gApp.fTimerID)
	g_source_remove(gApp.fTimerID);
      if (t)
	gApp.fTimerID = g_timeout_add(t,&MinWinIdleProc,null);
    }
  }

  int main(char[][] args) {
    gApp.cmdLineArgs = args;

    char*[] argv;
    argv.length = args.length;
    foreach (int n, char[] str; args) {
      argv[n] = toStringz(str);
    }
    int argc = args.length;
    char**argvp = argv.ptr;
    gtk_init(&argc,&argvp);

    int res = MinWinMain(&gApp);
    version(LOG) log.close();
    return res;
  }
}
