/* MinWin Group class
 * 
 * A Group is a lightweight component (meaning it has
 * no peer) that contains and lays out other components.
 *
 * A GroupBox is a potentially heavyweight component that
 * puts a box and label around other components.
 *
 * Written by Ben Hinkle and released to the public domain, as
 * explained at http://creativecommons.org/licenses/publicdomain
 * Email comments and bug reports to ben.hinkle@gmail.com
 */

module minwin.group;

private {
  import minwin.component;
  import minwin.window;
  import minwin.peerimpl;
  import minwin.logging;
  import std.string;
}

// lightweight group of other children - no peer
class Group : Component {
  Rect bounds;
  this(Component parent, char[] name = "") {
     hasPeer = NO_PEER;
     this.name = name;
     if (parent)
       parent.addChild(this);
   }
  void disposePeer(){}
  PeerForAdd getPeerForAdd() {
     return parent.getPeerForAdd();
   }
  void getPeerOffset(inout int x, inout int y) {
    int px, py;
    if (parent)
      parent.getPeerOffset(px,py);
    x = bounds.left+px;
    y = bounds.top+py;
  }
  void getBounds(inout Rect r) {
    r = bounds;
  }
  void setBounds(inout Rect r) {
    bounds = r;
    version(LOG) log.writefln("setting group bounds to %d %d %d %d",
    			      r.left,r.top,r.width,r.height);
    childLayoutDirty = true;
    repaint();
  }
  void size(Point s) {
    bounds.LTWH(bounds.left,bounds.top,s.x,s.y);
    childLayoutDirty = true;
    repaint();
  }
  void visible(bool vis) {
    foreach(Component ch; this) {
      ch.visible = vis;
    }
    childLayoutDirty = true;
    layout(false);
  }
  bool visible() {
    bool vis = false;
    foreach(Component ch; this) {
      vis = vis || ch.visible();
    }
    return vis;
  }
}
private import minwin.layout;
unittest {
  Group g = new Group(null);
  g.layoutMgr = new FlowLayout;
  Group g2 = new Group(g);
  g2.layoutMgr = new FlowLayout;
  g2.userPreferredSize(40,50);
  Group g3 = new Group(g);
  g3.layoutMgr = new FlowLayout;
  Group g4 = new Group(g3);
  g4.userPreferredSize(50,10);
  Group g5 = new Group(g3);
  g5.userPreferredSize(60,20);
  Group g6 = new Group(g3);
  g6.userPreferredSize(70,30);
  g.pack();
  Rect r2;
  g.getBounds(r2);
  assert( r2 == LTWH(0,0,70,110) );

  g6.getBounds(r2);
  assert( r2 == LTWH(0,30,70,30) );
  g5.getBounds(r2);
  assert( r2 == LTWH(5,10,60,20) );
  g4.getBounds(r2);
  assert( r2 == LTWH(10,0,50,10) );
  
  int x,y;
  g6.getPeerOffset(x,y);
  assert( x == 0 );
  assert( y == 80 );
}

version (Windows) {

  private {
    import minwin.app;
    import minwin.mswindows;
  }

  class GroupBox : WindowChild {
    int labelHeight;

    char[] text_data;
    char[] text() { return text_data; }
    void text(char[] c) {
      text_data = c;
      SendMessageA(peer,WM_SETTEXT,0,cast(LPARAM)toStringz(c));
    }

    this(Component parent, char[] text, char[] name = "") {
      PeerForAdd parentp = parent.getPeerForAdd();
      text_data = text;
      peer = CreateWindowA("BUTTON",toStringz(text), 
			   BS_GROUPBOX | WS_CHILD | WS_VISIBLE,
			   0,0,10,10,parentp,
			   cast(HMENU)0,gApp.hInstance,null);
      sysAssert(peer !is null, "Failed to create peer Button");
      setWindowChildPeer(this,peer,OWNS_PEER);
      this.name = name;
      Font f = standardFont(StandardFont.Gui);
      SendMessageA(peer,WM_SETFONT,cast(WPARAM)f.peer,0);
      HDC dc = GetDC(peer);
      HFONT oldFont = SelectObject(dc,f.peer);
      TEXTMETRICA metrics;
      GetTextMetricsA(dc,&metrics);
      labelHeight = metrics.tmHeight;
      SelectObject(dc,oldFont);
      ReleaseDC(peer,dc);
      parent.addChild(this);
    }


    void getLayoutBounds(inout Rect r) {
      super.getLayoutBounds(r);
      r.LTWH(r.left+2,r.top+labelHeight+2,
	      r.width-4,r.height-labelHeight-4);
    }

    Point preferredSize() {
      Point s;
      if (layoutMgr) {
	s = layoutMgr.preferredSize(this);
      }
      s.y = s.y + labelHeight+4;
      s.x = s.x + 4;
      if (userPreferredWidth() > 0)
	s.x = userPreferredWidth();
      if (userPreferredHeight() > 0)
	s.y = userPreferredHeight();
      return s;
    }
    mixin WindowChildImpl!();
  }

} else version (Motif) {

  private import minwin.motif;

  class GroupBox : WindowChild {
    Widget content;

    this(Component parent, char[] text, char[] name = "") {
      PeerForAdd parentp = parent.getPeerForAdd();
      this.name = name;
      char* str = toStringz(text);
      XmString xmstr = XmStringCreate(str,XmFONTLIST_DEFAULT_TAG);
      Arg[6] args;
      int n;
      //      XtSetArg(args[n],XmNlabelString,cast(XtArgVal)xmstr);n++;
      XtSetArg(args[n],XmNchildVerticalAlignment,cast(XtArgVal)XmALIGNMENT_CENTER);n++;
      peer = XmCreateFrame(parentp,toStringz(name),args,n);
      n = 0;
      XtSetArg(args[n],XmNlabelString,cast(XtArgVal)xmstr);n++;
      XtSetArg(args[n],XmNchildType,cast(XtArgVal)XmFRAME_TITLE_CHILD);n++;
      Widget label = XmCreateLabel(peer,"FrameLabel",args,n);
      XmStringFree(xmstr);
      setWindowChildPeer(this,peer,OWNS_PEER);
      parent.addChild(this);

      // add BulletinBoard to contain group children. The bboard sets
      // the Frame's size so we have to resize it by hand. 
      n = 0;
      XtSetArg(args[n],XmCAllowOverlap,cast(XtArgVal)true); n++;
      content = XmCreateBulletinBoard(peer,"Bulletin",args,n);
      XtAddEventHandler(peer,StructureNotifyMask,false,&mw_groupbox_layout_callback,this);

      XtManageChild(peer);
      XtRealizeWidget(peer);
      XtManageChild(label);
      XtRealizeWidget(label);
      XtManageChild(content);
      XtRealizeWidget(content);

    }

    void layout(bool validateParent) {
      Rect r = bounds;
      // TODO: don't fill up the whole frame since that overlaps the
      // border and title
      XtConfigureWidget(content,0,0,r.width,r.height,0);
      super.layout(validateParent);
    }

    // handle layout! Does Motif need a custom peer like GTK?
    PeerForAdd getPeerForAdd() { return content; }

    mixin WindowChildImpl!();
  }

  extern (C)
    void mw_groupbox_layout_callback(Widget w, XtPointer ud, 
				     XEvent* event, int* cont) {
    GroupBox box = cast(GroupBox) ud;
    if (box) {
      version(LOG)log.printf("in GroupBox.ConfigureNotify for %x\n",box);
      box.childLayoutDirty = true;
      box.layout(false);
      version(LOG)log.printf("in GroupBox.ConfigureNotify done layout foe %x\n",box);
    }
  }

} else version (GTK) {

  private import minwin.gtk;
  private import minwin.gtk_peers;

  class GroupBox : WindowChild {

    MinWinGtkPeer* content;

    char[] text() { 
      char* str = gtk_frame_get_label(cast(GtkFrame*)peer);
      if (str is null)
	return null;
      else
	return str[0..strlen(str)].dup;
    }
    void text(char[] c) {
      gtk_frame_set_label(cast(GtkFrame*)peer,g_strdup(c.ptr));
    }

    this(Component parent, char[] text, char[] name = "") {
      PeerForAdd parentp = parent.getPeerForAdd();
      this.name = name;
      char* str = toStringz(text);
      peer = gtk_frame_new(str);
      gtk_container_add(cast(GtkContainer*)parentp,peer);

      // add our peer to hook into GTK size allocation algorithm
      GtkWidget* wcontent = MinWinGtkPeer_new();
      content = cast(MinWinGtkPeer*)wcontent;
      gtk_widget_set_sensitive(wcontent,true);
      content.sizeRequest = &gtkRequest;
      content.sizeAllocate = &gtkAllocate;
      gtk_container_add(cast(GtkContainer*)peer,wcontent);

      setWindowChildPeer(this,peer,OWNS_PEER);
      parent.addChild(this);

      gtk_widget_realize(wcontent);
      gtk_widget_show(wcontent);
      gtk_widget_realize(peer);
      visible = true;
    }

    mixin WindowChildImpl!();

    PeerForAdd getPeerForAdd() {
      return cast(GtkWidget*)content;
    }

    // callback from the GTK peer asking for preferred size
    private void gtkRequest(GtkWidget *w, GtkRequisition* req) {
      version(LOG)log.writefln("gtkRequest for groupbox");
      Point s;
      if (layoutMgr) {
	s = layoutMgr.preferredSize(this);
	req.width = s.x;
	req.height = s.y;
      } else {
	req.width = 0;
	req.height = 0;
      }
      version(LOG)log.writefln("done gtkRequest for groupbox");
    }

    // callback from GTK peer telling us our position
    private void gtkAllocate(GtkWidget *w, GtkAllocation* req) {
      version(LOG)log.writefln("gtkAllocate for groupbox");
      if (layoutMgr) {
	layoutMgr.layout(this);
      }
      childLayoutDirty = false;
      version(LOG)log.writefln("done gtkAllocate for groupbox");
    }

    void getLayoutBounds(inout Rect r) {
      GtkWidget* w = cast(GtkWidget*)content;
      r = toRect(w.allocation);
    }

  }

}

